var labelType, useGradients, nativeTextSupport, animate;

///////////////////////////////////////////////////
//	le main code
///////////////////////////////////////////////////

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
    this.elem.style.left = (500 - this.elem.offsetWidth / 2) + 'px';
  }
};

		
function Listener(){
}
Listener.prototype.onCall = function(node, eventInfo, e, listenerFunction){
	// do nothing
}

function GraphFlow(){
	
	/////////////////////////////////////////////
	//				VARIABLES				   //
	/////////////////////////////////////////////
	
	//var that = this;
	
	/** Color Palettes */
	var nodeFillColor = [];
		nodeFillColor['Filter'] 	 	= "#DEDEDE";
		nodeFillColor['Repository']  	= "#EFAC6A";
		nodeFillColor['Reader'] 	 	= "#6AEF73";
	
	var nodeStrokeColor = [];
		nodeStrokeColor['Filter'] 	   	  = "#4D4D4D";
		nodeStrokeColor['Repository']  	  = "#735332";
		nodeStrokeColor['Reader'] 	      = "#327337";
		nodeStrokeColor['inputPort']   	  = "#AA0000";
		nodeStrokeColor['outputPort']	  = "#AA0000";
		nodeStrokeColor['repositoryPort'] = "#AA0000";
		
	var nodeColorFocus	  = "#0098BE",
		edgeColor 		  = "#114270",
		edgeColorFocus 	  = "#0098BE";
		
	/** Overall size modifier: the width of node labels in pixels */
	var vertexLabelSize 	= 12;

	/** FlowGraph Object */
	var fd;

	/** Nodes and edges where the mouse moves over */
	var hover = null;

	/** This array stores all Nodes and is initialized with a dummy,
		used for drawing edges between a node and the mousepointer.
		*/
	var json = [ {
			  "adjacencies": [], 
			  "data": {
				"$dim": 0,
				"$type": "none",
				"&xPos": 0,
				"&typeSelected" : null,
				"&yPos": 0,
			  }, 
			  "id": "#DUMMY_MOUSE_NODE", 
			  "name": "",
			  "alpha": 0
		}];

	/** selected node:	{id : <the id of the selected node, 
						 from : <the data.$type of the selected node>}
						 
		When clicking a node or edge, this variable will contain node
		information, that is required for creating edges.
		*/
	var selectedNode = null;
	
	/** This points to the DOM-node (not the json-array node!) that is
		an invisible dummy for moving an edge with the mouse
		*/
	var mouseNode = null;

	/** true if the mouse is over a port-node/edge */
	var mouseOverNode = false;
	var mouseOverEdge = false;
	
	/** A hashtable of mouseEvents to arrays of functions that
		should happen on these events
		*/
	var listener = [];
	
	/////////////////////////////////////////////
	//				FUNCTIONS				   //
	/////////////////////////////////////////////

	
	/**
	Sets the mouseCursor to a new one. If the new cursor is null,
	it will be reset to the default mouse cursor.
	*/
	this.setMouseCursor = function(newCursor){
		if(newCursor == null){
			fd.canvas.getElement().style.cursor = '';
		}else{
			fd.canvas.getElement().style.cursor = newCursor;
		}
	}
	
	this.graphToDot = function(filename){
		var node;
		var dotGraph = 'digraph structs {\nnode [shape=plaintext];\n\n'
		for(var n=0; n < json.length; n++){
			node = json[n];
			if(node.data.$type == "nodeFamily"){
				dotGraph += nodeToDot(node);
				dotGraph += '\n';
			}
		}
		dotGraph += '}';
		console.log(dotGraph);
	}
	
	/**
		Converts a single node to a node in dot language, and returns it as a string.
		@param nodeFamily - the node which is to be converted
	*/
	this.nodeToDot = function(nodeFamily){
		var pWidth = (nodeFamily.data.$width+vertexLabelSize)/2;
		var tableDef = '<TABLE BORDER="0" CELLSPACING="0" CELLBORDER="1" CELLPADDING="0">\n',
			portDefA = '<TR><TD WIDTH="'+pWidth+'" FIXEDSIZE="true" HEIGHT="'+(2*vertexLabelSize)+'" PORT="',
			portDefB = '">p</TD></TR>\n',
			padDefA = '<TR><TD FIXEDSIZE="true" WIDTH="'+pWidth+'" HEIGHT="',
			padDefB = '"></TD></TR>\n',
			emptyDef = '<TD FIXEDSIZE="true" HEIGHT="1" WIDTH="'+pWidth+'"></TD>\n';
		
		// this variable will contain a string, representing a node in dot-code
		var dot = nodeFamily.id,
			dotEdges = '';
		
		dot+= ' [label=< '+tableDef;
		dot+= '<TR><TD COLSPAN="2" HEIGHT="'+(2.5*vertexLabelSize)+'"></TD></TR><TR>\n';
		
		var portIndex = nodeFamily.data.$jsonIndex+2,
			lastPortIndex = nodeFamily.data.$jsonIndex+1+ nodeFamily.data.$portCount;
		var port = json[portIndex];
		if(port.data.$type == "inputPort"){
			
			dot += '<TD>'+tableDef;
			var dotInput = '',
				countInput = 0;
			
			// write the dot representation of ports into a temporary string
			while(portIndex <= lastPortIndex && port.data.$type == "inputPort"){
				dotInput += portDefA+ port.id + portDefB;
				countInput++;
				portIndex++;
				port = json[portIndex];
			}
			
			// apply padding if there is space above and below the ports
			var padding = nodeFamily.data.$height/2 - (1.5+countInput)*vertexLabelSize;
			if(padding > 0){
				dotInput = padDefA+padding+padDefB + dotInput + padDefA+padding+padDefB;
			}
			
			// append temp string to result
			dot += dotInput + '</TABLE></TD>';
		}
		else{
			dot += emptyDef;
		}
		
		if (portIndex <= lastPortIndex){
			// we get here if there were no input ports OR if we handled them all
			var dotRight = '',
				countRight = 0,
				adja,			// outgoing edges of port
				adjaIter;		// used to iterate through the edges
				
			// add repository ports
			if(port.data.$type == "repositoryPort"){
				
				dot += '<TD>'+tableDef;
				
				// write the dot representation of ports into a temporary string
				while(portIndex <= lastPortIndex && port.data.$type == "repositoryPort"){
					dotRight += portDefA+ port.id + portDefB;
					
					// add edges
					adja = port.adjacencies;
					for(a = 0; a < adja.length; a++){
						adjaIter = adja[a];
						dotEdges += nodeFamily.id+":"+port.id + ":e->" + adjaIter.data.$targetFamily + ":" + adjaIter.nodeTo+":w;\n";
					}
					
					countRight++;
					portIndex++;
					port = json[portIndex];
				}
			}
			// add output ports
			if(portIndex <= lastPortIndex && port.data.$type == "outputPort"){
				
				// if we have preceding repo ports, insert a little gap to separate
				// the output ports
				if(json[portIndex-1].data.$type == "repositoryPort"){
					dotRight += padDefA + vertexLabelSize + padDefB;
					countRight += 0.5;
				}
				else{
					dot += '<TD>'+tableDef;
				}
				
				// write the dot representation of ports into a temporary string
				while(portIndex <= lastPortIndex && port.data.$type == "outputPort"){
					dotRight += portDefA+ port.id + portDefB;
					
					// add edges
					adja = port.adjacencies;
					for(var a = 0; a < adja.length; a++){
						adjaIter = adja[a];
						dotEdges += nodeFamily.id+":"+port.id + ":e->" + adjaIter.data.$targetFamily + ":" + adjaIter.nodeTo+":w;\n";
					}
					
					countRight++;
					portIndex++;
					port = json[portIndex];
				}
			} 
			// apply padding  to right sideif there is space above and below the ports
				var padding = nodeFamily.data.$height/2 - (1.5+countRight)*vertexLabelSize;
				if(padding > 0){
					dotRight = padDefA+padding+padDefB + dotRight + padDefA+padding+padDefB;
				}
				dot += dotRight + '</TABLE></TD>';
		}else{
			dot += emptyDef;
		}
		// end of ports
		
		dot+= '</TR><TR><TD COLSPAN="2" HEIGHT="'+vertexLabelSize/2+'"></TD></TR></TABLE>>];\n';
		dot+= dotEdges;
		
		return dot;
	}
	
	/** 
		reloads the graph, trying to restore all positions 
		this function needs to be called whenever nodes or egdges are
		added or deleted!
	*/
	this.refresh = function(){
		// reload graph
		fd.loadJSON(json);
		// restore old node positions
		restoreNodePositions();
		fd.plot();
	}
	
	/**
		Adds a functionality to the following MouseEvents:
			onClick,
			onRightClick,
			onMouseMove,
			onMouseEnter,
			onMouseLeave,
			onDragStart,
			onDragMove,
			onDragCancel,
			onDragEnd
		Or one of these events:
			onCreateEdge(sourceNodeID, targetNodeID, sourcePortID, targetPortID),
			onCreateNode(nodeID),
			onRemoveEdge(sourceNodeID, targetNodeID, sourcePortID, targetPortID),
			onRemoveNode(nodeID)
		When altering the graph, do not forget to refresh() it at the
		end!
		@param event - the name of the mouse event ( see list above)
		@param listenerFunction - a function that may alter the graph
	*/
	this.addListener = function(event, listenerFunction){
		if(listener[event] == undefined){
			listener[event] = [];
		}
		listener[event].push(listenerFunction);
	}
	
	/**
	Adds a listener that forbids edge creation if...
		... an edge with the same source and target exists.
		...	the target node is the source node.
		... the source and target node are of the same port type.
		... a repository node is not connected to a repository and
			vice versa.
	*/
	this.addEdgeConstraints = function(){
		addListener("onCreateEdge", 
		function(sourceFamilyID, targetFamilyID, sourcePortID, targetPortID){
		
			// remove Edge if it leads to itself
			if( sourcePortID == targetPortID){
				removeEdge(sourcePortID, targetPortID);
				return;
			}
		
			var	sourcePort = getNode(sourcePortID),
				targetPort = getNode(targetPortID);
			
			// remove Edge if the connected nodes share the same type
			if(sourcePort.data.$type == targetPort.data.$type){
				removeEdge(sourcePortID, targetPortID);
				return;
			}
			
			var target = getNode(targetFamilyID),
				goesToRepo = (target.data.$nodeType == "Repository"),
				isRepoPort = (sourcePort.data.$type == "repositoryPort");
			
			// remove Edge if it is falsely connected to repository ports
			if(goesToRepo != isRepoPort){
				removeEdge(sourcePortID, targetPortID);
				return;
			}
			
			// remove duplicate Edge
			var adja = sourcePort.adjacencies,
				duplicate = false;
			for(var a=0; a < adja.length; a++){
				if(adja[a].nodeTo == targetPortID){
					if(duplicate){
						removeEdge(sourcePortID, targetPortID);
						return;
					}
					else{
						duplicate = true;
					}
				}
			}
		});
	}
	
	/**
	Changes all colors of the graph. Colors that should not be changed are given null as argument.
	*/
	/*
	this.setColors = function(nodeFocus, edgeFocus, edge, filterFill, filterStroke, readerFill, readerStroke, repoFill, repoSroke){
		
		var nodeFillColor = [];
		nodeFillColor['Filter'] 	 	= "#DEDEDE";
		nodeFillColor['Repository']  	= "#EFAC6A";
		nodeFillColor['Reader'] 	 	= "#6AEF73";
	
	var nodeStrokeColor = [];
		nodeStrokeColor['Filter'] 	   	  = "#4D4D4D";
		nodeStrokeColor['Repository']  	  = "#735332";
		nodeStrokeColor['Reader'] 	      = "#327337";
		nodeStrokeColor['inputPort']   	  = "#AA0000";
		nodeStrokeColor['outputPort']	  = "#AA0000";
		nodeStrokeColor['repositoryPort'] = "#AA0000";
		
	var nodeColorFocus	  = "#0098BE",
		edgeColor 		  = "#114270",
		edgeColorFocus 	  = "#0098BE";
		
	}*/
	
	/**
		Calls all listener that are registered under the eventName.
		@param eventName - see 'addListener()'
		@param arguments - an array of arguments. For mouseEvents it is [node, eventInfo, e],
						   for nodeEvents it is [nodeID], 
						   and for edgeEvents it is [sourceNodeID, targetNodeID, sourcePortID, targetPortID]
	*/
	function callListener(eventName, arguments){
		//Log.write("call "+eventName+" "+arguments.length + " ("+Math.random()+")");
		var lArr = listener[eventName];
		if(lArr == undefined){
			return;
		}
		// iterate through all listener this event concerns
		if(arguments.length == 3){
			for(var l=0; l < lArr.length; l++){
				lArr[l](arguments[0], arguments[1], arguments[2]);
			}
		}else if(arguments.length == 1){
			for(var l=0; l < lArr.length; l++){
				lArr[l](arguments[0]);
			}
		}else if(arguments.length == 4){
			for(var l=0; l < lArr.length; l++){
				lArr[l](arguments[0], arguments[1], arguments[2], arguments[3]);
			}
		}
	}
	
	/**
	 Moves all nodes to their respective positions, which are stored within
	 their data. We need this function to dynamically remove graph elements
	 without losing the node positions.
	*/
	function restoreNodePositions(){
		for (var n = 0; n < json.length; n++){
			var x = json[n].data.$xPos,
				y = json[n].data.$yPos,
				node =  fd.graph.getNode(json[n].id);
			node.pos.setc(x,y);
		}
	}
	
	/**
	Adds a repository node to the graph.
	@param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	@param outputPort - properties of the single output port(id, name, tooltip)
	*/
	this.addRepository = function(xPosition, yPosition, nodeFamily, inputPort){
		addNode(xPosition, yPosition, nodeFamily, null, [inputPort], null, "Repository");
	}
	
	/**
	Adds a reader node to the graph.
	@param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	@param repositoryPorts - an array of properties for the repository ports (id, name, tooltip)
	@param outputPorts - an array of properties for the output ports (id, name, tooltip)
	*/
	this.addReader = function(xPosition, yPosition, nodeFamily, repositoryPorts, outputPorts){
		addNode(xPosition, yPosition, nodeFamily, repositoryPorts, null, outputPorts, "Reader");
	}
	
	/**
	Adds a filter node to the graph.
	@param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	@param outputPort - properties of the single output port(id, name, tooltip)
	@param repositoryPorts - an array of properties for the repository ports (id, name, tooltip)
	@param inputPorts - an array of properties for the input ports (id, name, tooltip)
	@param outputPorts - an array of properties for the output ports (id, name, tooltip)
	*/
	this.addFilter = function(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts){
		addNode(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts, "Filter");
	}
	
	/**
	 Adds a Node to the graph.
	 @param nodeFamily - describes some properties of the node: (id, name, nodeClass, tooltip)
	 @param inputPortIDs - an array of properties for the input ports (id, name, tooltip)
	 @param outputPortIDs - an array of properties for the output ports (id, name, tooltip)
	 */
	this.addNode = function(xPosition, yPosition, nodeFamily, repositoryPorts, inputPorts, outputPorts, nodeType){
	
	var countRepo = 0,
		countInput = 0,
		countOutput = 0;
		
	if(repositoryPorts != null){
		countRepo= repositoryPorts.length;
	}
	if(inputPorts != null){
		countInput= inputPorts.length;
	}
	if(outputPorts != null){
		countOutput= outputPorts.length;
	}
	
	var maxPorts = Math.max(countInput, countRepo+countOutput+0.5),
		size = vertexLabelSize*2;
	
	// insert some space between repo ports and output ports, only if both are present
	if(maxPorts > countInput && (countRepo == 0 || countOutput == 0)){
		maxPorts-= 0.5;
	}
		
	var width = (nodeFamily.name.length+4)*vertexLabelSize,
		height = size + 
			Math.max( 2*size, size/2 + maxPorts * size);
	
	// the classname may be longer than the diplayed name
	if( nodeFamily.nodeClass != undefined){
		width = Math.max(width,  (nodeFamily.nodeClass.length+6)*(vertexLabelSize-2));
	}
	
	// add big node box
	// change node color, depending on its type
	var nodeColor = nodeFillColor[nodeType],
		strokeColor = nodeStrokeColor[nodeType];
	var newNode = {
		  "data": {
			"$dim": size,					
			"$type": "nodeFamily",
			"$jsonIndex": json.length,
			"$portCount": countRepo+countInput+countOutput,
			"$width": width,
			"$height": height,
			"$color": strokeColor,
			"$fillColor": nodeColor,
			"$nodeClass": nodeFamily.nodeClass,
			"$nodeType": nodeType,
			"$xPos": xPosition,
			"$yPos": yPosition,
			"$tooltip": nodeFamily.tooltip
		  }, 
		  "id": nodeFamily.id, 
		  "name": nodeFamily.name
	};
	json.push(newNode);
	
	// add closeButton
	var closeButton ={
		  "adjacencies": [], 
		  "data": {
			"$dim": size,					
			"$type": "crossBox",
			"$family": newNode,
			"$xPos": xPosition+ width/2,
			"$yPos": yPosition- height/2,
			"$color": strokeColor,
			"$fillColor": nodeColor,
		   }, 
		  "id": nodeFamily.id+"_close", 
		  "name": "x"
	};
	json.push(closeButton);
	
	// Loop Info: 
	// This loop adds all port types to the temporary port array.
	// For fast position computation, the relative position
	// of the first port to the nodeFamily-box is computed for each port type
	// it is stored in relativeX, relativeY
	
	for(var portType = 0; portType < 3; portType++){
		var x, y, relativeX, relativeY,
			loopPorts, loopType;
		switch(portType){
			case 0: // input ports
				relativeX = -width/2;
				relativeY = (0.75 - countInput/2 )*size;
				loopPorts = inputPorts;
				loopType = "inputPort";
				break;
		
			case 1: // repository ports
				var repoSpace = 0.5;
				if(countOutput == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput+countRepo+repoSpace)/2 )*size;
				loopPorts = repositoryPorts;
				loopType = "repositoryPort";
				break;
				
			case 2: // output ports
				var repoSpace = 0.5;
				if(countRepo == 0){
					repoSpace = 0;
				}
				relativeX = width/2;
				relativeY = (0.75 - (countOutput-countRepo-repoSpace)/2 )*size;
				loopPorts = outputPorts;
				loopType = "outputPort";
				break;
		}
		
		// continue only if ports exist
		if(loopPorts != null && loopPorts.length > 0){
		
			x = xPosition+ relativeX;
			y = yPosition+ relativeY;
			
			for(var p=0; p < loopPorts.length; p++){
				var port = loopPorts[p];
				var newPort ={
					  "adjacencies": [], 
					  "data": {
						"$dim": size,					
						"$type": loopType,
						"$xPos": x,
						"$yPos": y,
						"$color": nodeStrokeColor[loopType],
						"$tooltip": port.tooltip}, 
					  "id": nodeFamily.id+"_"+port.id, 
					  "name": port.name};
					  
				if(p == 0){
					newPort.data.$relativeX= relativeX;
					newPort.data.$relativeY= relativeY;
				}
				json.push(newPort);
				y += size;
			}
		}
	}
	
	// call listener
	callListener("onCreateNode", [nodeFamily.id]);
	}

	/**
		Removes a Node and all its edges from the graph.
		@param nodeFamily - the node which is to be removed
	*/
	this.removeNode =  function(nodeFamily){
		
		var deleteFrom = nodeFamily.data.$jsonIndex, 
			deleteUntil = deleteFrom+ nodeFamily.data.$portCount+1,
			familyID = nodeFamily.id;
		
		// delete nodeFamily and closeButton
		delete json[deleteFrom];
		delete json[deleteFrom+1];
		
		// delete ports
		for(var n = deleteFrom+2;  n <= deleteUntil;  n++){
			var portID = json[n].id;
			
			// remove all incoming edges
			for(var e = 0;  e < json.length;  e++){
				var incNode = json[e];
				if(incNode != undefined){
					removeEdge(incNode.id, portID);	
				}
			}
			
			delete json[n];
		}
		
		// collapse json array
		var deletionSum = deleteUntil-deleteFrom+1;
		json.splice(deleteFrom, deletionSum);
		selectedNode = null;
		
		// update nodeFamily indices
		for(var n = deleteUntil-deletionSum+1;  n < json.length;  n++){
			var familyIndex = json[n].data.$jsonIndex;
			if(familyIndex != undefined){
				json[n].data.$jsonIndex = n;
			}
		}
		
		// call listener
		callListener("onRemoveNode", [familyID]);
		
		return;
	 }
	 
	/**
		Moves a node to the specified position.
		@param nodeFamily - the node which is to be moved
		@param x - desired x-Position
		@param y - desired y-Position
	*/
	 function moveNode(nodeFamily, x, y){
		var moveFrom = nodeFamily.data.$jsonIndex, 
			moveUntil = moveFrom+ nodeFamily.data.$portCount+1,
			dim = nodeFamily.data.$dim;
			
		// set pos of family box
		nodeFamily.pos.setc(x, y);
		json[moveFrom].data.$xPos = x;
		json[moveFrom].data.$yPos = y;
		moveFrom++;
		
		// set pos of close button
		var node = fd.graph.getNode(json[moveFrom].id),
			nodeX = x+ nodeFamily.data.$width/2,
			nodeY =  y- nodeFamily.data.$height/2;
			
		node.pos.setc(nodeX, nodeY);
		json[moveFrom].data.$xPos = nodeX;
		json[moveFrom].data.$yPos = nodeY;
		moveFrom++;
		
		// set pos of ports
		for(var i= moveFrom; i<= moveUntil; i++){
			node = fd.graph.getNode(json[i].id);
			
			if(node.data.$relativeX != undefined){
				nodeX = x+ node.data.$relativeX;
				nodeY = y+ node.data.$relativeY;
			}
			node.pos.setc(nodeX,nodeY);
			json[i].data.$xPos = nodeX;
			json[i].data.$yPos = nodeY;
			nodeY+= dim;
		}
	 }

  
  /**
   *  adds an edge to the graph
   *  @param sourceID the id of the node from which the edge starts
   *  @param targetID the id of the node where the edge ends
   *  @return true if the edge was successfully added
   */
	this.addEdge = function(sourceID, targetID){
		
		// look up the source node and the nodeFamilys of both nodes
		var source, sourceFamilyID, targetFamilyID, lastFamilyID;
		for(var n=0; n< json.length; n++){
			var loopNode = json[n];
			if(loopNode.data.$type == "nodeFamily"){
				lastFamilyID = loopNode.id
			}
			else if(loopNode.id == sourceID){
				source = loopNode;
				sourceFamilyID = lastFamilyID;
			}
			else if(loopNode.id == targetID){
				targetFamilyID = lastFamilyID;
			}
		}
		
		var adja = source.adjacencies;
		// source node does not exist?
		if(adja == undefined){
			Log.write("Error: "+sourceID+" does not exist!");
			return false;
		}
	
		var edge ={
				  "nodeTo": targetID, 
				  "nodeFrom": sourceID,
				  "data": {"$direction" : [ sourceID, targetID ],
						   "$targetFamily" : targetFamilyID}
				  };
				  
		adja.push(edge);
		
		if(sourceID == mouseNode.id || targetID == mouseNode.id){
			edge.data.$type = "mouseArrow";
			return true;
		}
		
		// call listener (only if the edge is not connected to the dummy node!)
		callListener("onCreateEdge", [sourceFamilyID, targetFamilyID, sourceID, targetID]);
		
		return true;
	}
  
  /**
	Removes the edge between two vertices.
	@param source the source Node
	@param targetID the id of the target Node
	*/
  this.removeEdge = function(sourceID, targetID){
	var source = getNode(sourceID),
		adja = source.adjacencies;
	
	// nothing to delete?
	if(adja == undefined){
		return;
	}
	
	// look up nodeFamily IDs
	var sourceFamilyID, targetFamilyID, lastFamilyID;
		for(var n=0; n< json.length; n++){
			var loopNode = json[n];
			if(loopNode != undefined){ // <- this happens on Node Deletion
				if(loopNode.data.$type == "nodeFamily"){
					lastFamilyID = loopNode.id
				}
				else if(loopNode.id == sourceID){
					sourceFamilyID = lastFamilyID;
				}
				else if(loopNode.id == targetID){
					targetFamilyID = lastFamilyID;
				}
			}
		}
	
	for(var a=0; a< adja.length; a++){
		if(adja[a].nodeTo == targetID){
			delete adja[a];
			adja.splice(a,1);
			break;
		}
	}
	
	// call listener
	if(sourceID != mouseNode.id && targetID != mouseNode.id){
		callListener("onRemoveEdge", [sourceFamilyID, targetFamilyID, sourceID, targetID]);
	}
  }
  
  /**
	makes it so that a text label is not selectable.
	(unfortunately, this does not prevent CTRL+A selection)
	*/
  function setLabelUnselectable(label){
	// internet explorer:
	if (typeof label.onselectstart!="undefined"){
        label.onselectstart =
			function(){return false};
		}
	// mozilla firefox:
    else if (typeof label.style.MozUserSelect!="undefined"){
        label.style.MozUserSelect = "none";
	}
	// others:
    else{
        label.onmousedown =
			function(){return false};
	}
  }
  
    /**
	saves the given x, y position of a node
	@param nodeID - the id of the target node
	@param x - x-position of the node
	@param y - y-position of the node
	*/
  function saveNodePosition(nodeID, x, y){
	var node = getNode(nodeID);
	node.data.$xPos = x;
	node.data.$yPos = y;
  }
  
  /**
  returns a jsonNode by id
  @param nodeID the id of the desired node
  */
  this.getNode = function(nodeID){
	for(var n=0; n< json.length; n++){
		var node = json[n];
		if(node != undefined && node.id == nodeID){
			return json[n];
		}
	}
	return null;
  }
  
  /**
	clears all highlights, caused by mouse-enter events
	@param clearNodes - if true, highlighted nodes will be cleared
	@param clearEdges - if true, highlighted edges will be cleared
	*/
  function setHighlight(node){
		// animation parameters
		var trans = $jit.Trans.Quint.easeOut, 
			dur= 200;
			
		// clean up the old highlight
		if(hover != null){
			
			if(hover.nodeFrom != undefined){
				hover.setData('lineWidth', 1, 'end');
				hover.setData('color', edgeColor, 'end');
			}
			else{
				var type = hover.data.$type;
				if(type == "nodeFamily"){
					hover.setData('color', nodeStrokeColor[hover.data.$nodeType], 'end');
				}
				else if(type == "crossBox"){
					hover.setData('color', nodeStrokeColor[hover.data.$family.data.$nodeType], 'end');
				}
				else{
					hover.setData('color', nodeStrokeColor[type], 'end');
				}
			}
			hover = null;
		}
		
		// do we highlight something new?
		if(node != null){
			if(node.nodeFrom != undefined){
				node.setData('color', edgeColorFocus, 'end');
				node.setData('lineWidth', 2, 'end');
			}else{
				node.setData('color', nodeColorFocus, 'end');
			}
			hover = node;
		}else{
			// slower fade away transition
			trans = $jit.Trans.Quint.easeIn;
			dur = 400;
		}
		
		// Animations overwrite each other. If we let this animation
		// execute while we delete an edge, the onComplete() function
		// of the edge animation will not be called
		fd.fx.animate({  
				modes: ['edge-property:lineWidth:color',
						'node-property:color'],
				transition: trans,  
				duration: dur
		}); 
  }
    
  
  /**
	Initializes the Graph
	*/
  this.initGraph = function(graph){
  
	  if(graph != null){
		json = graph;
	  }
	  
	  fd = new $jit.FlowGraph({
		//id of the visualization container
		injectInto: 'infovis',
		//Enable zooming and panning
		//with scrolling and DnD
		Navigation: {
		  enable: true,
		  //Enable panning events only if we're dragging the empty
		  //canvas (and not a node).
		  panning: 'avoid nodes',
		  zooming: 10 //zoom speed. higher is more sensible
		},
		// Change node and edge styles such as
		// color and width.
		// These properties are also set per node
		// with dollar prefixed data-properties in the
		// JSON structure.
		Node: {
		  overridable: true,
		  //color: "#FFFFFF",
		  CanvasStyles: {  
			  lineWidth: 2  
			}  
		  
		},
		Edge: {
		  overridable: true,
		  color: edgeColor,
		  lineWidth: 1,
		  dim: vertexLabelSize,
		  type: 'flowArrow'
		},
		//Native canvas text styling  
		Label: {  
			type: 'HTML', //Native or HTML  
			size: 10,  
			style: 'bold'  
		}, 
		//Add Tips  
		  Tips: {  
			enable: true,  
			type: 'Native',
			onShow: function(tip, node) {  
			  var message = node.data.$tooltip;
			  
			  // do not show a tooltip if it is undefined
			  if(message == undefined){
				tip.style.visibility = "hidden";
			  }
			  else{
				tip.innerHTML = "<div class=\"tip-text\">" + message + "</div>";  
				tip.style.visibility = "visible";
			  }
			}  
		 },  
		// Add node events
		Events: {
		  enable: true,
		  enableForEdges: true,
		  type: 'Native',
		  
		  /**
		   *	EVENT:	OnRightClick
		   */
		  onRightClick: function(node, eventInfo, e) {
			callListener("onRightClick", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragEnd
		   */
		  onDragEnd: function(node, eventInfo, e) {
			callListener("onDragEnd", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragCancel
		   */
		  onDragCancel: function(node, eventInfo, e) {
			callListener("onDragCancel", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseEnter
		   */
		  onMouseEnter: function(node, eventInfo, e) {
			if(node.id == mouseNode.id){
				return;
			}
			
			// ignore some highlights if we are dragging an edge
			if(selectedNode != null && 
				(node.data.$type != "inputPort" 
				&& node.data.$type != "outputPort" 
				&& node.data.$type != "repositoryPort")){
				return;
			}
			
			// check if object is an edge or a node
			mouseOverNode = (node.nodeFrom == undefined);
			mouseOverEdge = !mouseOverNode;
			
			setHighlight(node);
			
			if(node.data.$type == "nodeFamily"){
				fd.canvas.getElement().style.cursor = 'move';
			}else{
				fd.canvas.getElement().style.cursor = 'pointer';
			}
			callListener("onMouseEnter", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseMove
		   */
		  onMouseMove: function(node, eventInfo, e) {
			// are we dragging an edge with the mouse ?
			if(selectedNode != null){
				var pos = eventInfo.getPos();
				mouseNode.pos.setc(pos.x, pos.y-2);
				fd.plot();
			}
			callListener("onMouseMove", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnMouseLeave
		   */
		  onMouseLeave: function(node, eventInfo, e) {
			fd.canvas.getElement().style.cursor = '';
			mouseOverNode = false;
			mouseOverEdge = false;
			
			setHighlight(null);
			callListener("onMouseLeave", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragStart
		   */
		  onDragStart: function(node, eventInfo, e) {
			if(node == false || node.id == mouseNode.id ||
			   node.data.$type != "nodeFamily"){
				return;
			}
			
			// memorize where we drag the node instead on centering
			// it to the mouse pointer
			var mousePos = eventInfo.getPos(),
				nodePos = node.pos.getc(true);
				
			mouseNode.data.$dragX = nodePos.x - mousePos.x;
			mouseNode.data.$dragY = nodePos.y - mousePos.y;
			callListener("onDragStart", [node, eventInfo, e]);
		  },
		  
		  /**
		   *	EVENT:	OnDragMove
		   */
		  onDragMove: function(node, eventInfo, e) {
			if(node == false || node.id == mouseNode.id ||
			   node.data.$type != "nodeFamily"){
				return;
			}
			var pos = eventInfo.getPos();
			
			moveNode(node, 
						   pos.x+mouseNode.data.$dragX,
							pos.y+mouseNode.data.$dragY);
			callListener("onDragMove", [node, eventInfo, e]);
			fd.plot();
			
		  },
		  
		  /**
		   *	EVENT:	OnTouchMove
		   */
		  onTouchMove: function(node, eventInfo, e) {
			$jit.util.event.stop(e); //stop default touchmove event
			this.onDragMove(node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnClick
		   */
		  onClick: function(node, eventInfo, e){
		  var mousePos = eventInfo.getPos();
		  
		  if(mouseOverNode && node != false){
		  
			switch(node.data.$type){
				case "crossBox":
					removeNode(node.data.$family);
					refresh();
					break;
					
				case "nodeFamily":
					break;
					
				case "outputPort":
				case "inputPort":
				case "repositoryPort":
					// no selection yet
					if(selectedNode == null){
						selectedNode = {"id": node.id, "from": node.data.$type};
						
						// add edge to mouseNode
						if(selectedNode.from == "inputPort"){
							addEdge(mouseNode.id, node.id);
						}
						else{
							addEdge(node.id, mouseNode.id);
						}
						refresh();
						return;
					}
					// click outside the edge drag box
					else{
						// add Edge if the selectedNode differs from the clickedNode
						var newEdgeAdded;
						if(selectedNode.from == "inputPort"){
							newEdgeAdded = addEdge(node.id, selectedNode.id);
						}
						else{
							newEdgeAdded = addEdge(selectedNode.id, node.id);
						}
						if(newEdgeAdded){
							refresh();
						}
					}
					break;
			}
			
		  }
		  else if(mouseOverEdge){
			// no selection yet
				if(selectedNode == null){
					selectedNode = {"id":node.data.$direction[0], "from":"outputPort"};//nodeTo
					
					// remove selectedEdge
					removeEdge(selectedNode.id, node.data.$direction[1]);
					
					// add edge to mouseNode
					addEdge(selectedNode.id, mouseNode.id);
					refresh();
					return;
				}
		  }
		  
		  // remove selection and mouse-edge upon clicking anywhere
		  if(selectedNode != null){
		  
			if(selectedNode.from == "inputPort"){
				removeEdge(mouseNode.id, selectedNode.id);
			}else{
				removeEdge(selectedNode.id, mouseNode.id);
			}
			
			selectedNode = null;
			refresh();
		  }
		  callListener("onClick", [node, eventInfo, e]);
		  }
		},
		//Number of iterations for the FD algorithm
		iterations: 0,
		//Edge length
		levelDistance: 2*vertexLabelSize,
		// This method is only triggered
		// on label creation and only for DOM labels (not native canvas ones).
		onCreateLabel: function(domElement, node){
			// do not add labels if it is a dummy element
			// instead save pointer to domElement
			if(node.id == "#DUMMY_MOUSE_NODE"){
				mouseNode = node;
				return;
			}
			if(node.data.$type != "nodeFamily"){
				return;
			}
			
			// save node position
			saveNodePosition(node.id, node.pos.x, node.pos.y);
			
			domElement.innerHTML = node.name;  
			
			// set style for name
			var style = domElement.style;  
			style.fontSize = 1.65* fd.canvas.scaleOffsetX* vertexLabelSize+"px";
			style.left = 0.2*vertexLabelSize-domElement.offsetWidth/2+"px";
			setLabelUnselectable(domElement);
		},
		
		// Change node styles when DOM labels are placed
		// or moved.
		onPlaceLabel: function(domElement, node){
		  if(node.data.$type != "nodeFamily"){
			return;
		  }
		  
		  var style = domElement.style;
		  var left = parseInt(style.left);
		  var top = parseInt(style.top);
		  var scale = fd.canvas.scaleOffsetX;
		  var w = scale*vertexLabelSize* node.name.length;//domElement.offsetWidth;
		  var h = scale*vertexLabelSize*1.65;//domElement.offsetHeight;
		  
		  // TODO: store more calculations inside a nodefamily!
		  var classWidth =  scale*(vertexLabelSize-2)* (node.data.$nodeClass.length+2);
		  if(classWidth > w){
			w = classWidth;
		  }
		  
		  style.fontSize = scale* 1.65 * vertexLabelSize+"px";
		  style.fontFamily = "Lucida Console";
		  style.left = (left - w/ 2) + 'px';
		  style.top = (top - h ) + 'px';
		  style.color = node.data.$color;
		  style.display = '';
		  // add class name
			if(node.data.$nodeClass != undefined){
				domElement.innerHTML = node.name;  
				var fontSize= scale*1.65*(vertexLabelSize-2);
				domElement.innerHTML += "<div style=\"font-size:"+fontSize+"px\"> &lt"+node.data.$nodeClass+"&gt";
			}
		}
	  });
	  
	  // load JSON data.
	  fd.loadJSON(json,0);
	  // compute positions
	  refresh();
	  // end
	  //alert("final wut:"+fd.nodeSizeModifier);
	}
	
	return this;
}

// test data
function init(){
	var addNodeCounter = 0;
	var graph = GraphFlow();
	// add a new listener: 
	graph.addListener("onRightClick", function(node,info,e){
										var newNode = {"id":"nuNode"+addNodeCounter, 
														"name":"node #"+addNodeCounter, 
														"nodeClass":"someClassAgain",
														"tooltip":"i am new and shiny!"};
										graph.addFilter(info.getPos().x, info.getPos().y, newNode, null,
													  [{"name":"inputPort","id":"ip1"}], 
													  [{"name":"outputPort", "id":"op1"}]); 
										graph.refresh();
										addNodeCounter++;
					  });
	/*graph.addListener("onRightClick", function(node, info, e){
			graph.graphToDot(null);
	});*/	
	
	// adds a listener that removes invalid Edges
	graph.addEdgeConstraints();
	
	//graph.addListener("onRemoveNode", function(nodeID){alert("bye bye, "+ nodeID+".");});
	
	// define graph by adding nodes
	var node1 = {"id":"superNode1", 
					  "name":"i am node", 
					  "nodeClass":"someFilter",
					  "tooltip":"look at me, i'm a node!"};
	
	graph.addFilter(0, -100, node1,[{"name":"repoPort", "id":"rp1", "tooltip":"repo"},
									{"name":"repoPort", "id":"rp2", "tooltip":"repo"},
									{"name":"repoPort", "id":"rp3", "tooltip":"repo"}],
	
								   [{"name":"inputPort","id":"ip1", "tooltip":"input goes here"}, 
									{"name":"inputPort","id":"ip2", "tooltip":"input goes here"}, 
								    {"name":"inputPort","id":"ip3", "tooltip":"input goes here"}, 
								    {"name":"inputPort","id":"ip4"}, 
								    {"name":"inputPort","id":"ip5", "tooltip":"input goes here"},
								    {"name":"inputPort","id":"ip6"}],
									
								   [{"name":"outputPort", "id":"op1"}]);

	// create graph
	graph.initGraph(null);
	
	// add nodes after graph creation
	var node2 = {"id":"superNode2", 
					  "name":"Super Repo", 
					  "nodeClass":"Reposiotry",
					  "tooltip":"look at me, i'm another node!"};
	graph.addRepository(100, 0, node2,{"name":"inputPort", "id":"ip1"});
	var node3 = {"id":"superNode3", 
					  "name":"i am reader", 
					  "nodeClass":"Reader",
					  "tooltip":"look at me, i'm another node!"};
	graph.addReader(0, 100, node3,[{"name":"repoPort", "id":"rp1"}], [{"name":"outputPort", "id":"op1"}]);
	// refresh graph to show the new nodes
	graph.refresh();
}
